@@ -39,7 +39,7 @@ require Exporter;
             STATUS_SERVER
             COA_REQUEST COA_ACCEPT COA_REJECT COA_ACK COA_NAK);
 
-$VERSION = '0.26';
+$VERSION = '0.31';
 
 my (%dict_id, %dict_name, %dict_val, %dict_vendor_id, %dict_vendor_name );
 my ($request_id) = $$ & 0xff;   # probably better than starting from 0
@@ -117,74 +117,80 @@ sub new {
         $serv_port = $SERVICES{$service};
     }
 
-    ($host, $port) = split(/:/, $h{'Host'});
-    if (!$port) {
-        $port = $serv_port;
-    }
-
     $self->{'timeout'} = $h{'TimeOut'} ? $h{'TimeOut'} : 5;
     $self->{'localaddr'} = $h{'LocalAddr'};
     $self->{'secret'} = $h{'Secret'};
     $self->{'message_auth'}  = $h{'Rfc3579MessageAuth'};
-    print STDERR "Using Radius server $host:$port\n" if $debug;
-    my %io_sock_args = (
-                Type => SOCK_DGRAM,
-                Proto => 'udp',
-                Timeout => $self->{'timeout'},
-                LocalAddr => $self->{'localaddr'},
-    );
+
     if ($h{'NodeList'}) {
         # contains resolved node list in text representation
         $self->{'node_list_a'} = {};
         foreach my $node_a (@{$h{'NodeList'}}) {
             my ($n_host, $n_port) = split(/:/, $node_a);
-            if (!$n_port) {
-                $n_port = $serv_port;
-            }
+            $n_port ||= $serv_port;
             my @hostinfo = gethostbyname($n_host);
             if (!scalar(@hostinfo)) {
                 print STDERR "Can't resolve node hostname '$n_host': $! - skipping it!\n" if $debug;
                 next;
             }
-            print STDERR "Adding ".inet_ntoa($hostinfo[4]).':'.$n_port." to node list.\n" if $debug;
+
+            my $ip = inet_ntoa($hostinfo[4]);
+            print STDERR "Adding ".$ip.':'.$n_port." to node list.\n" if $debug;
             # store split address to avoid additional parsing later
-            $self->{'node_list_a'}->{inet_ntoa($hostinfo[4]).':'.$n_port} =
-                    [inet_ntoa($hostinfo[4]), $n_port];
+            $self->{'node_list_a'}->{$ip.':'.$n_port} = [$ip, $n_port];
         }
+
         if (!scalar(keys %{$self->{'node_list_a'}})) {
             return $self->set_error('ESOCKETFAIL', 'Empty node list.');
         }
-        if ($host) {
-            my @hostinfo = gethostbyname($host);
-            if (scalar(@hostinfo)) {
-                my $act_addr_a = inet_ntoa($hostinfo[4]).':'.$port;
-                if (exists($self->{'node_list_a'}->{$act_addr_a})) {
-                    $self->{'node_addr_a'} = $act_addr_a;
-                } else {
-                    print STDERR "'$host' doesn't exist in node list - ignoring it!\n" if $debug;
-                }
-            } else {
-                print STDERR "Can't resolve active node hostname '$host': $! - ignoring it!\n" if $debug;
-            }
-        }
-    } else {
+    }
+
+    if ($h{'Host'}) {
+        ($host, $port) = split(/:/, $h{'Host'});
+        $port ||= $serv_port;
+        print STDERR "Using Radius server $host:$port\n" if $debug;
+
         my @hostinfo = gethostbyname($host);
         if (!scalar(@hostinfo)) {
+            if ($self->{'node_list_a'}) {
+                print STDERR "Can't resolve hostname '$host'\n" if $debug;
+                return $self;
+            }
+
             return $self->set_error('ESOCKETFAIL', "Can't resolve hostname '".$host."'.");
         }
-        $self->{'node_addr_a'} = inet_ntoa($hostinfo[4]).':'.$port;
-    }
-    if ($host) {
-        $io_sock_args{'PeerAddr'} = $host;
-        $io_sock_args{'PeerPort'} = $port;
+
+        my $ip = inet_ntoa($hostinfo[4]);
+
+        # if Host used with NodeList - it must be from the list
+        if ($self->{'node_list_a'} && !exists($self->{'node_list_a'}->{$ip.':'.$port})) {
+            print STDERR "'$host' doesn't exist in node list - ignoring it!\n" if $debug;
+            return $self;
+        }
+
+        # set as active node
+        $self->{'node_addr_a'} = $ip.':'.$port;
+
+        my %io_sock_args = (
+            Type => SOCK_DGRAM,
+            Proto => 'udp',
+            Timeout => $self->{'timeout'},
+            LocalAddr => $self->{'localaddr'},
+            PeerAddr => $host,
+            PeerPort => $port,
+        );
         $self->{'sock'} = IO::Socket::INET->new(%io_sock_args)
             or return $self->set_error('ESOCKETFAIL', $@);
     }
-    $self;
+
+    return $self;
 }
 
 sub send_packet {
     my ($self, $type, $retransmit) = @_;
+
+    $self->{attributes} //= '';
+
     my $data;
     my $length = 20 + length($self->{attributes});
 
@@ -377,10 +383,9 @@ sub clear_attributes {
 }
 
 sub _decode_enum {
-    my ( $name, $value, $format ) = @_;
+    my ( $name, $value) = @_;
 
-    $value = unpack( $format, $value );
-    if ( defined( $dict_val{$name}{$value} ) ) {
+    if ( defined $value && defined( $dict_val{$name}{$value} ) ) {
         $value = $dict_val{$name}{$value}{name};
     }
 
@@ -388,18 +393,48 @@ sub _decode_enum {
 }
 
 sub _decode_string {
-    my ( $self, $vendor, $id, $name, $value ) = @_;
+    my ( $self, $vendor, $id, $name, $value, $has_tag ) = @_;
 
     if ( $id == 2 && $vendor eq NO_VENDOR ) {
         return '<encrypted>';
     }
 
-    return $value;
+    if ($has_tag) {
+        my $tag = unpack('C', substr($value, 0, 1));
+        # rfc2868 section-3.3
+        # If the Tag field is greater than 0x1F, it SHOULD be
+        # interpreted as the first byte of the following String field.
+        if ($tag > 31) {
+            print STDERR "Attribute $name has tag value $tag bigger than 31 - ignoring it!\n" if $debug;
+            $tag = undef;
+        }
+        else {
+            # cut extracted tag
+            substr($value, 0, 1, '');
+        }
+        return ($value, $tag);
+    }
+
+    return ($value);
 }
 
 sub _decode_integer {
-    my ( $self, $vendor, $id, $name, $value ) = @_;
-    return _decode_enum( $name, $value, 'N' );
+    my ( $self, $vendor, $id, $name, $value, $has_tag ) = @_;
+
+    my $tag;
+    if ($has_tag) {
+        $tag = unpack('C', substr($value, 0, 1));
+        if ($tag > 31) {
+            print STDERR "Attribute $name has tag value $tag bigger than 31 - ignoring it!\n" if $debug;
+            $tag = undef;
+        }
+        else {
+            substr($value, 0, 1, "\x00");
+        }
+    }
+
+    $value = unpack('N', $value);
+    return (_decode_enum( $name, $value), $tag);
 }
 
 sub _decode_ipaddr {
@@ -476,6 +511,11 @@ sub _decode_sublist {
     return join( '; ', @values );
 }
 
+sub _decode_octets {
+    my ( $self, $vendor, $id, $name, $value ) = @_;
+    return '0x'.unpack("H*", $value);
+}
+
 my %decoder = (
     # RFC2865
     string  => \&_decode_string,
@@ -483,6 +523,7 @@ my %decoder = (
     ipaddr  => \&_decode_ipaddr,
     date    => \&_decode_integer,
     time    => \&_decode_integer,
+    octets  => \&_decode_octets,
     # RFC3162
     ipv6addr   => \&_decode_ipv6addr,
     ipv6prefix => \&_decode_ipv6prefix,
@@ -495,11 +536,12 @@ my %decoder = (
 );
 
 sub _decode_value {
-    my ( $self, $vendor, $id, $type, $name, $value ) = @_;
+    my ( $self, $vendor, $id, $type, $name, $value, $has_tag ) = @_;
 
     if ( defined $type ) {
         if ( exists $decoder{$type} ) {
-            return $decoder{$type}->( $self, $vendor, $id, $name, $value );
+            my ($decoded, $tag) = $decoder{$type}->( $self, $vendor, $id, $name, $value, $has_tag );
+            return wantarray ? ($decoded, $tag) : $decoded;
         }
         else {
             if ($debug) {
@@ -519,7 +561,7 @@ sub _decode_value {
 sub get_attributes {
     my $self = shift;
     my ( $vendor, $vendor_id, $name, $id, $length, $value, $type, $rawvalue, $tag, @a );
-    my ($attrs) = $self->{attributes};
+    my $attrs = $self->{attributes} // '';
 
     $self->set_error;
 
@@ -539,12 +581,8 @@ sub get_attributes {
 
         $name  = $r->{name} // $id;
         $type  = $r->{type};
-        $tag = undef;
-        if ($r->{has_tag}) {
-            ( $tag, $rawvalue ) = unpack('Ca*', $rawvalue);
-        }
 
-        $value = $self->_decode_value( $vendor, $id, $type, $name, $rawvalue );
+        ($value, $tag) = $self->_decode_value( $vendor, $id, $type, $name, $rawvalue, $r->{has_tag} );
 
         push(
             @a, {
@@ -567,12 +605,13 @@ sub vendorID ($) {
     my ($attr) = @_;
     if (defined $attr->{'Vendor'}) {
         return ($dict_vendor_name{ $attr->{'Vendor'} }{'id'} // int($attr->{'Vendor'}));
-    } else {
+    } elsif (exists $dict_name{$attr->{'Name'}} ) {
         # look up vendor by attribute name
         my $vendor_name = $dict_name{$attr->{'Name'}}{'vendor'} or return NO_VENDOR;
         my $vendor_id = $dict_vendor_name{$vendor_name}{'id'} or return NO_VENDOR;
         return $vendor_id;
     }
+    return NO_VENDOR;
 }
 
 sub _encode_enum {
@@ -586,7 +625,7 @@ sub _encode_enum {
 }
 
 sub _encode_string {
-    my ( $self, $vendor, $id, $name, $value ) = @_;
+    my ( $self, $vendor, $id, $name, $value, $tag ) = @_;
 
     if ( $id == 2 && $vendor eq NO_VENDOR ) {
         $self->gen_authenticator();
@@ -599,12 +638,21 @@ sub _encode_string {
     #   return pack('C', 0) . substr($_[0], 0, 246);
     # }
 
+    if (defined $tag) {
+        $value = pack('C', $tag) . $value;
+    }
+
     return $value;
 }
 
 sub _encode_integer {
-    my ( $self, $vendor, $id, $name, $value ) = @_;
-    return _encode_enum( $name, $value, 'N' );
+    my ( $self, $vendor, $id, $name, $value, $tag ) = @_;
+    $value = _encode_enum( $name, $value, 'N' );
+    if (defined $tag) {
+        # tag added to 1st byte, not extending the value length
+        substr($value, 0, 1, pack('C', $tag) );
+    }
+    return $value;
 }
 
 sub _encode_ipaddr {
@@ -784,11 +832,11 @@ my %encoder = (
 );
 
 sub _encode_value {
-    my ( $self, $vendor, $id, $type, $name, $value ) = @_;
+    my ( $self, $vendor, $id, $type, $name, $value, $tag ) = @_;
 
     if ( defined $type ) {
         if ( exists $encoder{$type} ) {
-            return $encoder{$type}->( $self, $vendor, $id, $name, $value );
+            return $encoder{$type}->( $self, $vendor, $id, $name, $value, $tag );
         }
         else {
             if ($debug) {
@@ -807,7 +855,7 @@ sub _encode_value {
 
 sub add_attributes {
     my ($self, @attr) = @_;
-    my ($a, $vendor, $id, $type, $value);
+    my ($a, $vendor, $id, $type, $value, $need_tag);
     my @a = ();
     $self->set_error;
 
@@ -822,6 +870,12 @@ sub add_attributes {
             $attr_name = $1;
         }
 
+        if (! exists $dict_name{$attr_name}) {
+            # no dictionaries loaded, $attr_name must be attribute ID
+            push @a, $attr;
+            next;
+        }
+
         $id = $dict_name{$attr_name}{id} // int($attr_name);
         $vendor = vendorID($attr);
         if (exists($dict_name{$attr_name}{'tlv'})) {
@@ -850,10 +904,21 @@ sub add_attributes {
     }
 
     for $a (@a) {
-        $id = $dict_name{ $a->{Name} }{id} // int($a->{Name});
-        $type = $a->{Type} // $dict_name{ $a->{Name} }{type};
-        $vendor = vendorID($a);
-        my $need_tag = (defined $a->{Tag}) || $dict_name{ $a->{Name} }{has_tag};
+        if (exists $dict_name{ $a->{Name} }) {
+            my $def = $dict_name{ $a->{Name} };
+            $id = $def->{id};
+            # allow to override Type (why?)
+            $type = $a->{Type} // $def->{type};
+            $need_tag = $a->{Tag} // $def->{has_tag};
+        }
+        else {
+            # ID must be a value for Name
+            $id = int($a->{Name});
+            $type = $a->{Type};
+            $need_tag = $a->{Tag};
+        }
+
+        # we do not support 0 value for Tag
         if ($need_tag) {
             $a->{Tag} //= 0;
             if ($a->{Tag} < 1 || $a->{Tag} > 31) {
@@ -862,32 +927,36 @@ sub add_attributes {
             }
         }
 
+        $vendor = vendorID($a);
         if ($vendor eq WIMAX_VENDOR) {
-            # WiMAX uses non-standard VSAs - include the continuation byte
+            #TODO WiMAX uses non-standard VSAs - include the continuation byte
         }
 
-        unless (defined($value = $self->_encode_value($vendor, $id, $type, $a->{Name}, $a->{Value}))) {
-            print STDERR "Unable to encode attribute $a->{Name} ($id, $type, $vendor) with value '$a->{Value}'\n" if $debug;
+        unless (defined($value = $self->_encode_value($vendor, $id, $type, $a->{Name}, $a->{Value}, $a->{Tag}))) {
+            printf STDERR "Unable to encode attribute %s (%s, %s, %s) with value '%s'\n",
+                $a->{Name}, $id // '?', $type // '?', $vendor, $a->{Value}
+            if $debug;
             next;
         }
-        print STDERR "Adding attribute $a->{Name} ($id, $type, $vendor) with value '$a->{Value}'\n" if $debug;
+
+        if ($debug) {
+            printf STDERR "Adding attribute %s (%s, %s, %s) with value '%s'%s\n",
+                    $a->{Name}, $id, $type, $vendor,
+                    $a->{Value},
+                    ($a->{Tag} ? sprintf(' (tag:%d)', $a->{Tag}) : '');
+        }
+
         if ( $vendor eq NO_VENDOR ) {
-            if ($need_tag) {
-                $self->{'attributes'} .= pack('C C C', $id, length($value) + 3, $a->{Tag} // 0) . $value;
-            }
-            else {
-                $self->{'attributes'} .= pack('C C', $id, length($value) + 2) . $value;
-            }
+            # tag already included in $value, if any
+            $self->{'attributes'} .= pack('C C', $id, length($value) + 2) . $value;
         } else {
             # VSA
             # pack vendor-ID + vendor-type + vendor-length
             if ($vendor eq WIMAX_VENDOR) {
                 # add continuation byte
                 $value = pack('N C C C', $vendor, $id, length($value) + 3, 0) . $value;
-            } elsif ($need_tag) {
-                # tagged attribute
-                $value = pack('N C C C', $vendor, $id, length($value) + 3, $a->{Tag} // 0) . $value;
             } else {
+                # tag already included in $value, if any
                 $value = pack('N C C', $vendor, $id, length($value) + 2) . $value;
             }
 
@@ -1046,6 +1115,11 @@ sub load_dictionary {
             $dict_vendor_id{$id}{'name'} = $name;
         } elsif ($cmd eq 'begin-vendor') {
             $dict_def_vendor = $name;
+            if (! $freeradius_dict) {
+                # force format
+                $freeradius_dict = 1;
+                print STDERR "Detected BEGIN-VENDOR, switch to FreeRADIUS dictionary format\n" if $debug;
+            }
         } elsif ($cmd eq 'end-vendor') {
             $dict_def_vendor = NO_VENDOR;
         } elsif ($cmd eq 'begin-tlv') {
@@ -1060,7 +1134,7 @@ sub load_dictionary {
             my @path = split("/", $file);
             pop @path; # remove the filename at the end
             my $path = ( $name =~ /^\// ) ? $name : join("/", @path, $name);
-            load_dictionary('', $path);
+            load_dictionary('', $path, %opt);
         }
     }
     $fh->close;
@@ -1289,7 +1363,9 @@ was loaded). Values for C<TYPE> can be 'C<string>', 'C<integer>', 'C<ipaddr>',
 'C<ipv6addr>', 'C<ipv6prefix>', 'C<ifid>' or 'C<avpair>'. The C<VENDOR> may be
 Vendor's name from the dictionary or their integer id. For tagged attributes
 (RFC2868) tag can be specified in C<Name> using 'Name:Tag' format, or by
-using C<Tag> pair. TAG value is expected to be an integer.
+using C<Tag> pair. TAG value is expected to be an integer, within [1:31] range
+(zero value isn't supported).
+
 
 =item get_attributes
 
